           (*-------------------------------------------*
            |  The Dining Mathematicians in Csp-Prover  |
            |               August 2004                 |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory DM01_Sys_def = Csp_Prover:

(*****************************************************************

         1. defines Sys
         2. int lemmas
         3. 
         4. 

 *****************************************************************)

(*********************************************************
                    ODD and EVEN
 *********************************************************)

consts
  EVEN :: "int => bool"
  ODD  :: "int => bool"

defs
  EVEN_def : "EVEN n == (n mod 2 = 0)"
  ODD_def  : "ODD n  == (n mod 2 = 1)"

syntax
  "_EVENs" :: "int set" ("EVENs")
  "_ODDs"   :: "int set" ("ODDs")

translations
  "EVENs"  == "Collect EVEN"
  "ODDs"   == "Collect ODD"

(*********************************************************
                         event
 *********************************************************)

datatype Event = Eat0 | Back0 | End0 | RD0 int | WR0 int
               | Eat1 | Back1 | End1 | RD1 int | WR1 int 
               | NUM int | undef 

lemma expand_Event_fun[simp]:
  "ALL Ef Eg. ((Ef::int => Event) = Eg) = (ALL n. Ef n = Eg n)"
apply (simp add: expand_fun_eq)
done
                
syntax
  "_CH0" :: "Event set" ("CH0")
  "_CH1" :: "Event set" ("CH1")
  "_OBS" :: "Event set" ("OBS")

translations
  "CH0" == "(range RD0) Un (range WR0)"
  "CH1" == "(range RD1) Un (range WR1)"
  "OBS" == "{Eat0, Back0, End0, Eat1, Back1, End1}"

(*********************************************************
                     function
 *********************************************************)

consts
  getInt :: "Event => int"

primrec
  "getInt (Eat0)   = 0"
  "getInt (Eat1)   = 0"
  "getInt (Back0)  = 0"
  "getInt (Back1)  = 0"
  "getInt (End0)   = 0"
  "getInt (End1)   = 0"
  "getInt (RD0 n)  = n"
  "getInt (RD1 n)  = n"
  "getInt (WR0 n)  = n"
  "getInt (WR1 n)  = n"
  "getInt (undef)  = 0"
  "getInt (NUM n)  = n"

(*********************************************************
                Parallel system definition
 *********************************************************)

datatype SysName = VAR int | TH0 | EAT0 int | TH1 | EAT1 int

consts 
  SysDef :: "(SysName, Event) procDef"

primrec
  "SysDef      (TH0) = RD0 ? n -> IF (EVEN n)
                                  THEN Eat0  -> <EAT0 n>
                                  ELSE Back0 -> <TH0>"

  "SysDef      (TH1) = RD1 ? n -> IF (ODD n)
                                  THEN Eat1  -> <EAT1 n>
                                  ELSE Back1 -> <TH1>"

  "SysDef   (EAT0 n) = End0 -> WR0 ! (n div 2) -> <TH0>"

  "SysDef   (EAT1 n) = End1 -> WR1 ! (3 * n + 1) -> <TH1>"

  "SysDef    (VAR n) = WR0 ? n -> <VAR n>
                   [+] WR1 ? n -> <VAR n>
                   [+] RD0 ! n -> <VAR n>
                   [+] RD1 ! n -> <VAR n>"

consts
  Sys :: "int => (SysName, Event) procRec"

defs
  Sys_def:
   "Sys  == (%n.
     LET SysDef
     IN (<TH0> |[CH0]| <VAR n> |[CH1]| <TH1>) -- (CH0 Un CH1))"

(*********************************************************
         To unfold "range" and "syntactic sugar", ...
 *********************************************************)

declare image_iff  [simp]
declare csp_ss_def [simp]

(*********************************************************
               nohide and guarded lemmas
 *********************************************************)

lemma guardSysDef[simp]: "!!C. guard ((SysDef) C)"
by (induct_tac C, simp_all)

lemma nohideSysDef[simp]: "!!C. nohide ((SysDef) C)"
by (induct_tac C, simp_all)

(*********************************************************
                        Lemmas
 *********************************************************)

(*** int lemmas ***)

lemma int_le_inc: "ALL (n::int) m. n <= m --> n = m | n+1 <= m"
apply (auto)
done

lemma mod_2_not_le: "ALL (n::int). ~(2 <= n mod 2)"
apply (simp add: linorder_not_le)
done

lemma mod_2_or: "ALL (n::int). n mod 2 = 0 | n mod 2 =1"
apply (intro allI)
apply (insert int_le_inc)
apply (drule_tac x="0" in spec)
apply (drule_tac x="n mod 2" in spec)
apply (simp)
apply (erule disjE)
apply (simp)

apply (insert int_le_inc)
apply (drule_tac x="1" in spec)
apply (drule_tac x="n mod 2" in spec)
apply (simp)
apply (erule disjE)
apply (simp)
apply (simp add: mod_2_not_le)
done

(*** ODD and EVEN lemmas ***)

lemma EVEN_not_ODD[simp]: "ALL n. EVEN n = (~ ODD n)"
apply (auto simp add: EVEN_def ODD_def )
apply (insert mod_2_or)
apply (drule_tac x="n" in spec)
apply (simp)
done

lemma ODD_add_1: "ALL n. (n mod 2 = 1) --> ((n + 1) mod 2 = (0::int))"
apply (simp add: zmod_zadd_left_eq)
done

lemma ODD_EX: "ALL m. ODD m --> (EX n. m = 2 * n + 1)"
apply (intro allI impI)
apply (simp add: ODD_def)
apply (insert ODD_add_1)
apply (drule_tac x="m" in spec)
apply (simp)
apply (simp add: zmod_eq_0_iff)
apply (erule exE)
apply (rule_tac x="q - 1" in exI)
apply (simp)
done

lemma ODD_to_EVEN[simp]: "ODD n ==> EVEN (3 * n + 1)"
apply (insert ODD_EX)
apply (drule_tac x="n" in spec)
apply (simp add: ODD_def EVEN_def)
apply (erule exE)
apply (simp add: zmod_zadd_left_eq)
apply (simp add: zmod_zmult1_eq')
done

lemma ODD_to_notODD[simp]: "ODD n ==> ~ ODD (3 * n + 1)"
apply (insert ODD_to_EVEN[of n])
apply (simp)
done

(*** range ***)

lemma fold_range: "(EX a. x = f a) = (x : range f)"
apply (auto)
done

(*********************************************************
    lemma for stable failures equivalece and refinement
 *********************************************************)

(*** unfold VAR ***)

lemma VAR:
     "LET SysDef IN <VAR n>
   =F[af] LET SysDef IN 
       ? x:insert (RD1 n) (insert (RD0 n) ((range WR0) Un (range WR1)))
         -> IF (x : (range WR0) | x : (range WR1)) THEN <VAR (getInt x)> ELSE <VAR n>"
apply (tactic {* csp_unwind_left_tac 1 *})
apply (tactic {* csp_step_left_tac 1 *})
apply (auto)
apply (tactic {* csp_simp_tac 1 *})+
done

lemmas unfold_SYS_rules = csp_rule_rw[OF VAR]
lemmas fold_SYS_rules   = csp_rule_rev_rw[OF VAR[simplified]]

(*** unfold TH0 ***)

lemma TH0:
     "LET SysDef IN <TH0>
   =F[af] LET SysDef IN RD0 ? n -> IF (EVEN n)
                                   THEN Eat0  -> <EAT0 n>
                                   ELSE Back0 -> <TH0>"
apply (tactic {* csp_unwind_left_tac 1 *})
done

lemmas unfold_SYS_rules = unfold_SYS_rules csp_rule_rw[OF TH0]
lemmas fold_SYS_rules   = fold_SYS_rules csp_rule_rev_rw[OF TH0[simplified]]

(*** unfold TH1 ***)

lemma TH1:
     "LET SysDef IN <TH1>
   =F[af] LET SysDef IN RD1 ? n -> IF (ODD n)
                                   THEN Eat1  -> <EAT1 n>
                                   ELSE Back1 -> <TH1>"
apply (tactic {* csp_unwind_left_tac 1 *})
done

lemmas unfold_SYS_rules = unfold_SYS_rules csp_rule_rw[OF TH1]
lemmas fold_SYS_rules   = fold_SYS_rules csp_rule_rev_rw[OF TH1[simplified]]

(*** unfold EAT0 ***)

lemma EAT0:
     "LET SysDef IN <EAT0 n>
   =F[af] LET SysDef IN ? x:{End0} -> WR0 ! (n div 2) -> <TH0>"
apply (tactic {* csp_unwind_left_tac 1 *})
done

lemmas unfold_SYS_rules = unfold_SYS_rules csp_rule_rw[OF EAT0]
lemmas fold_SYS_rules   = fold_SYS_rules csp_rule_rev_rw[OF EAT0[simplified]]

(*** unfold EAT1 ***)

lemma EAT1:
     "LET SysDef IN <EAT1 n>
   =F[af] LET SysDef IN ? x:{End1} -> WR1 ! (3 * n + 1) -> <TH1>"
apply (tactic {* csp_unwind_left_tac 1 *})
done

lemmas unfold_SYS_rules = unfold_SYS_rules csp_rule_rw[OF EAT1]
lemmas fold_SYS_rules   = fold_SYS_rules csp_rule_rev_rw[OF EAT1[simplified]]

(*** TH0 VAR step 1 ***)

lemma TH0_VAR: 
  "LET SysDef IN 
        (<TH0> |[CH0]| <VAR n>)
   =F
   LET SysDef IN 
        (? x:insert (RD0 n) (insert (RD1 n) (range WR1)) 
          -> IF (x : (range RD0) | x : (range WR0))
                THEN IF EVEN (getInt x) 
                     THEN (Eat0  -> <EAT0 n> |[CH0]| <VAR n>)
                     ELSE (Back0 -> <TH0> |[CH0]| <VAR n>)
                ELSE (<TH0> |[CH0]| <VAR (getInt x)>))"
apply (tactic {* csp_simp_add_tac "unfold_SYS_rules" 1 *})
apply (tactic {* csp_step_left_tac 1 *})
apply (case_tac "EVEN n")
apply (auto)
apply (tactic {* csp_simp_add_tac "fold_SYS_rules" 1 *})
done

lemmas unfold_SYS_rules = unfold_SYS_rules  csp_rule_rw[OF TH0_VAR]
lemmas fold_SYS_rules   = fold_SYS_rules    csp_rule_rev_rw[OF TH0_VAR[simplified]]

(*** TH0 VAR TH1 step 1 ***)

lemma TH0_VAR_TH1:
  "LET SysDef IN
        (<TH0> |[CH0]| <VAR n> |[CH1]| <TH1>)
   =F
   LET SysDef IN ? x:{RD0 n, RD1 n} 
      -> IF (x = RD0 n)
            THEN IF EVEN n
                 THEN (Eat0 -> <EAT0 n> |[CH0]| <VAR n> |[CH1]| <TH1>)
                 ELSE (Back0 -> <TH0>   |[CH0]| <VAR n> |[CH1]| <TH1>)
            ELSE IF ODD n
                 THEN (<TH0> |[CH0]| <VAR n> |[CH1]| Eat1 -> <EAT1 n>)
                 ELSE (<TH0> |[CH0]| <VAR n> |[CH1]| Back1 -> <TH1>)"
apply (tactic {* csp_simp_add_tac "unfold_SYS_rules" 1 *})
apply (tactic {* csp_step_left_tac 1 *})
apply (case_tac "EVEN n")
apply (auto)
apply (tactic {* csp_simp_add_tac "fold_SYS_rules" 1 *})+
done

lemmas unfold_SYS_rules = unfold_SYS_rules  csp_rule_rw[OF TH0_VAR_TH1]
lemmas fold_SYS_rules   = fold_SYS_rules    csp_rule_rev_rw[OF TH0_VAR_TH1[simplified]]

end
